import sys
from datetime import datetime

import tornado.web
from jet_bridge.utils.async_exec import as_future
from jet_bridge_base.db import get_connection
from jet_bridge_base.exceptions.request_error import RequestError
from jet_bridge_base.sentry import sentry_controller
from tornado import gen
from six.moves.urllib_parse import parse_qs

from jet_bridge_base.request import Request
from jet_bridge_base.responses.redirect import RedirectResponse
from jet_bridge_base.responses.template import TemplateResponse
from jet_bridge_base.status import HTTP_204_NO_CONTENT
from tornado.iostream import StreamClosedError


class BaseViewHandler(tornado.web.RequestHandler):
    view = None

    def request_headers(self):
        return {k.upper().replace('-', '_'): v for k, v in self.request.headers.items()}

    def request_files(self):
        def map_file(arg):
            key, files = arg
            file = files[0]
            return key, (file.filename, file.body)

        return dict(map(map_file, self.request.files.items()))

    def get_request(self):
        query_arguments = parse_qs(self.request.query, keep_blank_values=True)
        return Request(
            self.request.method.upper(),
            self.request.protocol,
            self.request.host,
            self.request.path,
            self.path_kwargs,
            self.request.uri,
            query_arguments,
            self.request_headers(),
            self.request.body,
            self.request.body_arguments,
            self.request_files(),
            self.request,
            self
        )

    def before_dispatch(self, request):
        self.view.before_dispatch(request)

    def after_dispatch(self, request):
        self.view.after_dispatch(request)

    def on_finish(self):
        self.view.on_finish()

    def set_default_headers(self):
        for name, value in self.view.default_headers().items():
            self.set_header(name, value)

    @gen.coroutine
    def write_response(self, response):
        try:
            if isinstance(response, RedirectResponse):
                self.redirect(response.url, status=response.status)
                raise gen.Return()

            for name, value in response.header_items():
                self.set_header(name, value)

            if response.status is not None:
                self.set_status(response.status)

            if isinstance(response, TemplateResponse):
                yield self.render(response.template, **(response.data or {}))
            else:
                yield self.finish(response.render())
        except StreamClosedError:
            pass

        raise gen.Return()

    @gen.coroutine
    def write_error(self, status_code, **kwargs):
        exc_type = exc = traceback = None

        if kwargs.get('exc_info'):
            exc_type, exc, traceback = kwargs['exc_info']
        else:
            exc = Exception()

        try:
            request = self.get_request()
        except RequestError as e:
            request = e.request

        response = self.view.error_response(request, exc_type, exc, traceback)
        yield self.write_response(response)
        raise gen.Return()

    def options(self, *args, **kwargs):
        self.set_status(HTTP_204_NO_CONTENT)
        self.finish()

    @gen.coroutine
    def dispatch(self, action, *args, **kwargs):
        request = self.get_request()

        connection = get_connection(request)
        if connection:
            connection['last_request'] = datetime.now()
            sentry_controller.set_context('Database connection', {
                'name': connection.get('name'),
                'project': connection.get('project'),
                'token': connection.get('token')
            })
        else:
            sentry_controller.set_context('Database connection', {})

        def execute():
            self.before_dispatch(request)

            try:
                result = self.view.dispatch(action, request, *args, **kwargs)
            finally:
                self.after_dispatch(request)

            return result

        try:
            response = yield as_future(execute)
            yield self.write_response(response)
        except Exception:
            exc_type, exc, traceback = sys.exc_info()
            response = self.view.error_response(request, exc_type, exc, traceback)
            yield self.write_response(response)
        finally:
            raise gen.Return()

    def get(self, *args, **kwargs):
        return self.dispatch('get', *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.dispatch('post', *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.dispatch('put', *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self.dispatch('patch', *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.dispatch('delete', *args, **kwargs)


def view_handler(cls):
    class ViewHandler(BaseViewHandler):
        view = cls()

    return ViewHandler



import os

from jet_bridge.utils.async_exec import pool_submit
from jet_bridge_base.configuration import Configuration
from jet_bridge_base.utils.common import get_random_string

from jet_bridge import settings, VERSION

from six.moves.http_cookies import Morsel

Morsel._reserved[str('samesite')] = str('SameSite')


class JetBridgeConfiguration(Configuration):

    def get_type(self):
        return 'jet_bridge'

    def get_version(self):
        return VERSION

    def get_settings(self):
        return {
            'DEBUG': settings.DEBUG,
            'READ_ONLY': settings.READ_ONLY,
            'AUTO_OPEN_REGISTER': settings.AUTO_OPEN_REGISTER,
            'CONFIG': settings.CONFIG,
            'WEB_BASE_URL': settings.WEB_BASE_URL,
            'API_BASE_URL': settings.API_BASE_URL,
            'PROJECT': settings.PROJECT,
            'TOKEN': settings.TOKEN,
            'ENVIRONMENT': settings.ENVIRONMENT,
            'CORS_HEADERS': settings.CORS_HEADERS,
            'BASE_URL': settings.BASE_URL,
            'JWT_VERIFY_KEY': settings.JWT_VERIFY_KEY,
            'BEARER_AUTH_KEY': settings.BEARER_AUTH_KEY,
            'ENVIRONMENT_TYPE': settings.ENVIRONMENT_TYPE,
            'BLACKLIST_HOSTS': settings.BLACKLIST_HOSTS,
            'DATABASE_ENGINE': settings.DATABASE_ENGINE,
            'DATABASE_HOST': settings.DATABASE_HOST,
            'DATABASE_PORT': settings.DATABASE_PORT,
            'DATABASE_USER': settings.DATABASE_USER,
            'DATABASE_PASSWORD': settings.DATABASE_PASSWORD,
            'DATABASE_NAME': settings.DATABASE_NAME,
            'DATABASE_EXTRA': settings.DATABASE_EXTRA,
            'DATABASE_CONNECTIONS': settings.CONNECTIONS,
            'DATABASE_CONNECTIONS_OVERFLOW': settings.CONNECTIONS_OVERFLOW,
            'DATABASE_ONLY': settings.DATABASE_ONLY,
            'DATABASE_EXCEPT': settings.DATABASE_EXCEPT,
            'DATABASE_MAX_TABLES': settings.DATABASE_MAX_TABLES,
            'DATABASE_SCHEMA': settings.DATABASE_SCHEMA,
            'DATABASE_SSL_CA': settings.DATABASE_SSL_CA,
            'DATABASE_SSL_CERT': settings.DATABASE_SSL_CERT,
            'DATABASE_SSL_KEY': settings.DATABASE_SSL_KEY,
            'DATABASE_SSH_HOST': settings.DATABASE_SSH_HOST,
            'DATABASE_SSH_PORT': settings.DATABASE_SSH_PORT,
            'DATABASE_SSH_USER': settings.DATABASE_SSH_USER,
            'DATABASE_SSH_PRIVATE_KEY': settings.DATABASE_SSH_PRIVATE_KEY,
            'COOKIE_SAMESITE': settings.COOKIE_SAMESITE,
            'COOKIE_SECURE': settings.COOKIE_SECURE,
            'COOKIE_DOMAIN': settings.COOKIE_DOMAIN,
            'COOKIE_COMPRESS': settings.COOKIE_COMPRESS,
            'STORE_PATH': settings.STORE_PATH,
            'CACHE_METADATA': settings.CACHE_METADATA,
            'CACHE_METADATA_PATH': settings.CACHE_METADATA_PATH,
            'CACHE_MODEL_DESCRIPTIONS': settings.CACHE_MODEL_DESCRIPTIONS,
            'SSO_APPLICATIONS': self.clean_sso_applications(settings.SSO_APPLICATIONS),
            'ALLOW_ORIGIN': settings.ALLOW_ORIGIN,
            'TRACK_DATABASES': settings.TRACK_DATABASES,
            'TRACK_DATABASES_ENDPOINT': settings.TRACK_DATABASES_ENDPOINT,
            'TRACK_DATABASES_AUTH': settings.TRACK_DATABASES_AUTH,
            'TRACK_MODELS_ENDPOINT': settings.TRACK_MODELS_ENDPOINT,
            'TRACK_MODELS_AUTH': settings.TRACK_MODELS_AUTH,
            'TRACK_QUERY_SLOW_TIME': settings.TRACK_QUERY_SLOW_TIME,
            'TRACK_QUERY_HIGH_MEMORY': settings.TRACK_QUERY_HIGH_MEMORY,
            'RELEASE_INACTIVE_GRAPHQL_SCHEMAS_TIMEOUT': settings.RELEASE_INACTIVE_GRAPHQL_SCHEMAS_TIMEOUT,
            'DISABLE_AUTH': settings.DISABLE_AUTH
        }

    def media_get_available_name(self, path):
        dir_name, file_name = os.path.split(path)
        file_root, file_ext = os.path.splitext(file_name)

        while os.path.exists(os.path.join(settings.MEDIA_ROOT, path)):
            path = os.path.join(dir_name, '%s_%s%s' % (file_root, get_random_string(7), file_ext))

        return path

    def media_exists(self, path):
        absolute_path = os.path.join(settings.MEDIA_ROOT, path)
        return os.path.exists(absolute_path)

    def media_listdir(self, path):
        absolute_path = os.path.join(settings.MEDIA_ROOT, path)
        directories = []
        files = []

        for dirpath, dirnames, filenames in os.walk(absolute_path):
            directories.extend(dirnames)
            files.extend(filenames)

        return directories, files

    def media_get_modified_time(self, path):
        absolute_path = os.path.join(settings.MEDIA_ROOT, path)
        return os.path.getmtime(absolute_path)

    def media_size(self, path):
        absolute_path = os.path.join(settings.MEDIA_ROOT, path)
        return os.path.getsize(absolute_path)

    def media_open(self, path, mode='rb'):
        return open(path, mode)

    def media_save(self, path, content):
        absolute_path = os.path.join(settings.MEDIA_ROOT, path)

        if not os.path.exists(os.path.dirname(absolute_path)):
            try:
                os.makedirs(os.path.dirname(absolute_path))
            except OSError:
                pass

        with open(absolute_path, 'wb') as f:
            f.write(content)

        return path

    def media_delete(self, path):
        absolute_path = os.path.join(settings.MEDIA_ROOT, path)
        os.remove(absolute_path)

    def media_url(self, path, request):
        if settings.MEDIA_BASE_URL:
            url = '{}{}'.format(settings.MEDIA_BASE_URL, path)
        else:
            url = request.protocol + "://" + request.host + '/media/' + path

        return url

    def session_get(self, request, name, default=None, decode=True, secure=True):
        if secure:
            value = request.original_handler.get_secure_cookie(name)
        else:
            value = request.original_handler.get_cookie(name)

        if value is None:
            return default
        elif decode and secure:
            return value.decode()
        else:
            return value

    def session_set(self, request, name, value, secure=True):
        if value is None:
            self.session_clear(request, name)
        elif secure:
            request.original_handler.set_secure_cookie(
                name,
                value,
                samesite=settings.COOKIE_SAMESITE,
                secure=settings.COOKIE_SECURE,
                domain=settings.COOKIE_DOMAIN
            )
        else:
            request.original_handler.set_cookie(
                name,
                value,
                samesite=settings.COOKIE_SAMESITE,
                secure=settings.COOKIE_SECURE,
                domain=settings.COOKIE_DOMAIN
            )

    def session_clear(self, request, name):
        request.original_handler.clear_cookie(name)

    def run_async(self, func, *args, **kwargs):
        pool_submit(func, *args, **kwargs)



import multiprocessing
from concurrent.futures import Executor, ThreadPoolExecutor

from tornado.concurrent import Future, chain_future
from tornado.ioloop import IOLoop


class _AsyncExecution:
    """Tiny wrapper around ThreadPoolExecutor. This class is not meant to be
    instantiated externally, but internally we just use it as a wrapper around
    ThreadPoolExecutor so we can control the pool size and make the
    `as_future` function public.
    """

    def __init__(self, max_workers=None):
        self._max_workers = (
            max_workers or multiprocessing.cpu_count()
        )  # type: int
        self._pool = None  # type: Optional[Executor]

    def init_pool(self):
        self._pool = ThreadPoolExecutor(max_workers=self._max_workers)

    def set_max_workers(self, count):
        if self._pool:
            self._pool.shutdown(wait=True)

        self._max_workers = count
        self.init_pool()

    def as_future(self, query):
        # concurrent.futures.Future is not compatible with the "new style"
        # asyncio Future, and awaiting on such "old-style" futures does not
        # work.
        #
        # tornado includes a `run_in_executor` function to help with this
        # problem, but it's only included in version 5+. Hence, we copy a
        # little bit of code here to handle this incompatibility.

        if not self._pool:
            self.init_pool()

        old_future = self._pool.submit(query)
        new_future = Future()  # type: Future

        IOLoop.current().add_future(
            old_future, lambda f: chain_future(f, new_future)
        )

        return new_future

    def submit(self, query, *args, **kwargs):
        if not self._pool:
            self.init_pool()

        self._pool.submit(query, *args, **kwargs)


_async_exec = _AsyncExecution()

as_future = _async_exec.as_future

set_max_workers = _async_exec.set_max_workers

pool_submit = _async_exec.submit



from __future__ import unicode_literals

import os

import six
from prompt_toolkit import prompt, print_formatted_text, HTML
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.validation import Validator
from sqlalchemy import create_engine

from jet_bridge import settings
from jet_bridge_base.db import build_engine_url

engines = [
    {
        'name': 'postgresql',
        'default_port': 5432
    },
    {
        'name': 'mysql',
        'default_port': 3306
    },
    {
        'name': 'oracle',
        'default_port': 1521
    },
    {
        'name': 'mssql+pyodbc',
        'default_port': 1433
    },
    {
        'name': 'bigquery'
    },
    {
        'name': 'snowflake'
    },
    {
        'name': 'sqlite'
    },
    {
        'name': 'none'
    }
]


def is_not_empty(text):
    return text is not None and text.strip() != ''


def database_engine_is_valid(text):
    return text in list(map(lambda x: x['name'], engines))


def port_is_valid(text):
    try:
        int(text)
        return True
    except ValueError as e:
        return False


def is_file_exists(text):
    return os.path.exists(text) and os.path.isfile(text)


promt_messages = 0


def create_config(config_not_set):
    print_formatted_text(HTML('<skyblue><b>===========================================</b></skyblue>'))

    if config_not_set:
        print_formatted_text(HTML('<skyblue><b>Configuration is not set</b></skyblue>'))
    else:
        print_formatted_text(HTML('<skyblue><b>Configuration</b></skyblue>'))

    print_formatted_text(HTML('<skyblue>You will be asked to enter settings</skyblue>'))
    print_formatted_text(HTML('<skyblue>and config file will be generated</skyblue>'))
    print_formatted_text(HTML('<skyblue><b>===========================================</b></skyblue>\n'))

    # session = PromptSession()
    html_completer = WordCompleter(list(map(lambda x: x['name'], engines)))

    def promt_message(message):
        global promt_messages
        promt_messages += 1
        return HTML('{}. {}\n> '.format(promt_messages, message))

    if 'project' not in settings.USE_DEFAULT_CONFIG:
        project = prompt(
            promt_message('<green><b>Enter your project unique name</b></green>'),
            default=settings.PROJECT
        )

        print_formatted_text('')
    else:
        project = settings.PROJECT

    if 'token' not in settings.USE_DEFAULT_CONFIG:
        token = prompt(
            promt_message('<green><b>Enter your Jet Bridge token</b></green>'),
            default=settings.TOKEN
        )

        print_formatted_text('')
    else:
        token = settings.TOKEN

    if 'address' not in settings.USE_DEFAULT_CONFIG:
        address = prompt(
            promt_message('<green><b>Which host to run Jet Bridge on?</b></green>\n<i>Default is {}</i>'.format('0.0.0.0 (any IP)')),
            default=settings.ADDRESS or '0.0.0.0'
        )

        print_formatted_text('')
    else:
        address = settings.ADDRESS or '0.0.0.0'

    port = prompt(
        promt_message('<green><b>Which port to run Jet Bridge on?</b></green>\n<i>Default is {}</i>'.format('8888')),
        validator=Validator.from_callable(
            port_is_valid,
            error_message='Incorrect port',
            move_cursor_to_end=True
        ),
        default='{0}'.format(settings.PORT) if settings.PORT else '8888'
    )

    print_formatted_text('')

    database_engine = prompt(
        promt_message('<green><b>Which database do you use?</b></green>\n<i>Should be one of: {}</i>'.format(
            ', '.join(map(lambda x: x['name'], engines)))),
        validator=Validator.from_callable(
            database_engine_is_valid,
            error_message='Unknown database engine',
            move_cursor_to_end=True
        ),
        completer=html_completer,
        validate_while_typing=False,
        default=settings.DATABASE_ENGINE or ''
    )

    database_name = None
    database_host = None
    database_port = None
    database_user = None
    database_password = None
    database_extra = None

    while True:
        print_formatted_text('')

        if database_engine == 'none':
            database_host = ''
            database_port = ''
            database_name = ''
            database_user = ''
            database_password = ''

            break
        elif database_engine == 'sqlite':
            prompts = 1

            database_name = prompt(
                promt_message('<green><b>Enter your database file path</b></green>'),
                validator=Validator.from_callable(
                    is_file_exists,
                    error_message='File does not exist on given path',
                    move_cursor_to_end=True
                ),
                default=database_name or settings.DATABASE_NAME or ''
            )

            database_host = ''
            database_port = ''
            database_user = ''
            database_password = ''
        elif database_engine == 'bigquery':
            prompts = 5

            database_name = prompt(
                promt_message('<green><b>Enter your BigQuery project name</b></green>'),
                validator=Validator.from_callable(
                    is_not_empty,
                    error_message='BigQuery project name is required',
                    move_cursor_to_end=True
                ),
                default=database_name or settings.DATABASE_NAME or ''
            )

            print_formatted_text('')

            database_password = prompt(
                promt_message('<green><b>Enter your service account key</b></green>'),
                default=database_password or settings.DATABASE_PASSWORD or ''
            )
        elif database_engine == 'snowflake':
            prompts = 5

            message = '<green><b>Enter your account name</b></green>'

            database_host = prompt(
                promt_message(message),
                validator=Validator.from_callable(
                    is_not_empty,
                    error_message='Account name is required',
                    move_cursor_to_end=True
                )
            )

            print_formatted_text('')

            database_name = prompt(
                promt_message('<green><b>Enter your database name</b></green>'),
                validator=Validator.from_callable(
                    is_not_empty,
                    error_message='Database name is required',
                    move_cursor_to_end=True
                ),
                default=database_name or settings.DATABASE_NAME or ''
            )

            print_formatted_text('')

            database_user = prompt(
                promt_message('<green><b>Enter your database user</b></green>'),
                validator=Validator.from_callable(
                    is_not_empty,
                    error_message='Database user is required',
                    move_cursor_to_end=True
                ),
                default=database_user or settings.DATABASE_USER or ''
            )

            print_formatted_text('')

            database_password = prompt(
                promt_message('<green><b>Enter your database password</b></green>'),
                default=database_password or settings.DATABASE_PASSWORD or ''
            )

            print_formatted_text('')

            warehouse_name = prompt(
                promt_message('<green><b>Enter your warehouse name</b></green>'),
                default='compute_wh'
            )

            database_extra = 'warehouse={}'.format(warehouse_name)
        else:
            prompts = 5

            default = database_host or settings.DATABASE_HOST or 'localhost'
            message = '<green><b>Enter your database host</b></green>\n<i>Default is {}</i>'.format('localhost')

            if settings.POSSIBLE_HOST:
                message += '\n<b>{}</b> should point to <b>{}</b> on Docker environment'.format(settings.POSSIBLE_HOST, 'localhost')

            database_host = prompt(
                promt_message(message),
                validator=Validator.from_callable(
                    is_not_empty,
                    error_message='Database host is required',
                    move_cursor_to_end=True
                ),
                default=default
            )

            print_formatted_text('')

            default_port = list(map(lambda x: six.text_type(x.get('default_port', '')), filter(
                lambda x: x['name'] == database_engine, engines
            )))[0]

            database_port = prompt(
                promt_message('<green><b>Enter your database port</b></green>'),
                validator=Validator.from_callable(
                    is_not_empty,
                    error_message='Database port is required',
                    move_cursor_to_end=True
                ),
                default=database_port or '{0}'.format(settings.DATABASE_PORT) if settings.DATABASE_PORT else '{0}'.format(default_port)
            )

            print_formatted_text('')

            database_name = prompt(
                promt_message('<green><b>Enter your database name</b></green>'),
                validator=Validator.from_callable(
                    is_not_empty,
                    error_message='Database name is required',
                    move_cursor_to_end=True
                ),
                default=database_name or settings.DATABASE_NAME or ''
            )

            print_formatted_text('')

            database_user = prompt(
                promt_message('<green><b>Enter your database user</b></green>'),
                validator=Validator.from_callable(
                    is_not_empty,
                    error_message='Database user is required',
                    move_cursor_to_end=True
                ),
                default=database_user or settings.DATABASE_USER or ''
            )

            print_formatted_text('')

            database_password = prompt(
                promt_message('<green><b>Enter your database password</b></green>'),
                default=database_password or settings.DATABASE_PASSWORD or ''
            )

        engine_url = build_engine_url({
            'engine': database_engine,
            'host': database_host,
            'port': database_port,
            'name': database_name,
            'user': database_user,
            'password': database_password,
            'extra': database_extra
        })

        if not engine_url:
            continue

        engine = create_engine(engine_url)

        try:
            connection = engine.connect()
            connection.close()
        except Exception as e:
            global promt_messages
            print_formatted_text('')
            print_formatted_text(HTML('<red><b>Connection failed:</b></red>'))
            print_formatted_text(HTML('<red>{}</red>'.format(e)))
            print_formatted_text('')
            print_formatted_text(HTML('<skyblue><b>Please try again</b></skyblue>'))
            promt_messages -= prompts
            continue

        print_formatted_text('')
        print_formatted_text(HTML('<skyblue><b>Database connected successfully!</b></skyblue>'))
        print_formatted_text('')

        break

    if 'config' not in settings.USE_DEFAULT_CONFIG:
        print_formatted_text('')

        config = prompt(
            promt_message('<green><b>Where to store config file?</b></green>\nDefault is jet.conf inside current working directory\n[{}]'.format(os.path.abspath(settings.DEFAULT_CONFIG_PATH))),
            default=settings.CONFIG or '{0}'.format(settings.DEFAULT_CONFIG_PATH)
        )
    else:
        config = settings.CONFIG or '{0}'.format(settings.DEFAULT_CONFIG_PATH)

    config_content = {
        'ADDRESS': address,
        'PORT': port,
        'CONFIG': config,
        'PROJECT': project,
        'TOKEN': token,
        'DATABASE_ENGINE': database_engine,
        'DATABASE_HOST': database_host,
        'DATABASE_PORT': database_port,
        'DATABASE_NAME': database_name,
        'DATABASE_USER': database_user,
        'DATABASE_PASSWORD': database_password
    }

    if settings.ENVIRONMENT:
        config_content['ENVIRONMENT'] = settings.ENVIRONMENT

    if settings.WEB_BASE_URL and settings.WEB_BASE_URL != settings.DEFAULT_WEB_BASE_URL:
        config_content['WEB_BASE_URL'] = settings.WEB_BASE_URL

    if settings.API_BASE_URL and settings.API_BASE_URL != settings.DEFAULT_API_BASE_URL:
        config_content['API_BASE_URL'] = settings.API_BASE_URL

    try:
        os.makedirs(os.path.dirname(config))
    except OSError:
        pass

    with open(config, 'w') as f:
        f.write('[JET]\n')
        for key, value in config_content.items():
            f.write('{}={}\n'.format(key, value))

    print_formatted_text('')
    print_formatted_text(HTML('<skyblue><b>===========================================</b></skyblue>'))
    print_formatted_text(HTML('<skyblue><b>Configuration file is created at:</b></skyblue>'))
    print_formatted_text(HTML('<skyblue>{}</skyblue>'.format(os.path.abspath(config))))
    print_formatted_text('')
    print_formatted_text(HTML('<skyblue>You can now start <b>Jet Bridge</b></skyblue>'))
    print_formatted_text(HTML('<skyblue><b>===========================================</b></skyblue>\n'))
